import{d as m,al as d,t as _,b as n,c as s,e as c,f as v,o as a,g as k,h as t,cl as y,p as g,cm as x,cn as N,F as C,a as h}from"./index-BjFilUfY.js";import{u as B}from"./usePageTitle-J4-vayug.js";const V=m({__name:"Notifications",setup(b){const r=d(),e=_(r.notifications.getNotifications),i=n(()=>e.response??[]),l=n(()=>e.executed&&i.value.length===0),u=n(()=>e.executed);return B("Notifications"),(F,o)=>{const f=v("p-layout-default");return a(),s(f,{class:"notifications"},{header:c(()=>[k(t(y))]),default:c(()=>[u.value?(a(),g(C,{key:0},[l.value?(a(),s(t(x),{key:0})):(a(),s(t(N),{key:1,notifications:i.value,onDelete:o[0]||(o[0]=p=>t(e).refresh()),onUpdate:o[1]||(o[1]=p=>t(e).refresh())},null,8,["notifications"]))],64)):h("",!0)]),_:1})}}});export{V as default};
//# sourceMappingURL=Notifications-_-aRmHf3.js.map
