import{d as v,b as y,o as i,p as u,g as e,e as t,F as C,v as x,h as n,cW as B,m as f,B as w,c as P,f as l,i as V,cX as A,cY as g,a5 as F}from"./index-BjFilUfY.js";import{u as M}from"./usePageTitle-J4-vayug.js";import{u as T}from"./usePrefectApi-DHJDxwOU.js";const j={class:"settings-block"},D=v({__name:"SettingsCodeBlock",props:{engineSettings:{}},setup(p){const s=p,a=y(()=>Object.entries(s.engineSettings));return(m,c)=>(i(),u("div",j,[e(n(B),{multiline:""},{default:t(()=>[(i(!0),u(C,null,x(a.value,(o,r)=>(i(),u("div",{key:r,class:"settings-block--code-line"},f(o[0])+": "+f(o[1]),1))),128))]),_:1})]))}}),L=v({__name:"Settings",async setup(p){let s,a;const m=[{text:"Settings"}],c=T(),[o,r]=([s,a]=w(()=>Promise.all([c.admin.getSettings(),c.admin.getVersion()])),s=await s,a(),s);return M("Settings"),(E,d)=>{const b=l("p-key-value"),S=l("p-theme-toggle"),_=l("p-label"),k=l("p-layout-default");return i(),P(k,{class:"settings"},{header:t(()=>[e(n(V),{crumbs:m},{actions:t(()=>[e(b,{class:"settings__version",label:"Version",value:n(r),alternate:""},null,8,["value"])]),_:1})]),default:t(()=>[e(_,{label:"Theme"},{default:t(()=>[e(S)]),_:1}),e(_,{label:"Color Mode",class:"settings__color-mode"},{default:t(()=>[e(n(A),{selected:n(g),"onUpdate:selected":d[0]||(d[0]=h=>F(g)?g.value=h:null)},null,8,["selected"])]),_:1}),e(_,{label:"Server Settings"},{default:t(()=>[e(D,{class:"settings__code-block","engine-settings":n(o)},null,8,["engine-settings"])]),_:1})]),_:1})}}});export{L as default};
//# sourceMappingURL=Settings-Yth39v93.js.map
