import{d as s,s as u,w as l,c,r as m,a as p,o as i,H as f,q as v,l as d,P as y}from"./index-BjFilUfY.js";const r=a=>{const{value:e,isCustom:o}=f();if(v(a.query)&&o.value)try{const t=d.map("SavedSearchFilter",e.value,"LocationQuery");return{...a,query:t}}catch(t){console.error(t)}return!0},h=s({expose:[],beforeRouteEnter:r,beforeRouteUpdate:r}),q=s({...h,__name:"FlowRunsPageWithDefaultFilter",props:{component:{type:Function}},setup(a){const e=a,o=u(null);function t(n){return y(n)}return l(e.component,()=>{t(e.component)?e.component().then(n=>{o.value=n.default}):o.value=e.component},{immediate:!0}),(n,C)=>o.value!==null?(i(),c(m(o.value),{key:0})):p("",!0)}});export{q as default};
//# sourceMappingURL=FlowRunsPageWithDefaultFilter-BoWwl7CY-QNUJO57M.js.map
