import{d as i,al as f,t as _,b as a,c as u,e as r,f as v,o,g as b,h as t,bT as w,p as y,bU as Q,bV as g,F as x,a as C}from"./index-BjFilUfY.js";import{u as V}from"./usePageTitle-J4-vayug.js";const W=i({__name:"WorkQueues",setup(h){const l=f(),p={interval:3e4},e=_(l.workQueues.getWorkQueues,[{}],p),n=a(()=>e.response??[]),c=a(()=>e.executed&&n.value.length==0),d=a(()=>e.executed);return V("Work Queues"),(q,s)=>{const m=v("p-layout-default");return o(),u(m,{class:"queues"},{header:r(()=>[b(t(w))]),default:r(()=>[d.value?(o(),y(x,{key:0},[c.value?(o(),u(t(Q),{key:0})):(o(),u(t(g),{key:1,"work-queues":n.value,onUpdate:s[0]||(s[0]=k=>t(e).refresh()),onDelete:s[1]||(s[1]=k=>t(e).refresh())},null,8,["work-queues"]))],64)):C("",!0)]),_:1})}}});export{W as default};
//# sourceMappingURL=WorkQueues-BDoVwSRy.js.map
