import{d as p,al as _,c as l,e as o,f as m,o as d,g as e,h as n,co as h,cp as C,bN as i,aX as c,_ as N}from"./index-BjFilUfY.js";import{u as y}from"./usePageTitle-J4-vayug.js";const x=p({__name:"NotificationCreate",setup(b){const r=_();async function s(t){try{await r.notifications.createNotification(t),i.push(c.notifications())}catch(a){N("Error creating notification","error"),console.warn(a)}}function f(){i.push(c.notifications())}return y("Create Notification"),(t,a)=>{const u=m("p-layout-default");return d(),l(u,{class:"notification-create"},{header:o(()=>[e(n(h))]),default:o(()=>[e(n(C),{action:"Create",onSubmit:s,onCancel:f})]),_:1})}}});export{x as default};
//# sourceMappingURL=NotificationCreate-BBgpTZfD.js.map
