import{d as m,al as p,t as d,b as s,c as a,e as c,f as _,o as t,g as f,h as o,c4 as k,p as i,c5 as y,c6 as B,F as h,a as x}from"./index-BjFilUfY.js";import{u as C}from"./usePageTitle-J4-vayug.js";const F=m({__name:"Blocks",setup(b){const n=p(),e=d(n.blockDocuments.getBlockDocumentsCount),l=s(()=>e.executed&&e.response==0),r=s(()=>e.executed);return C("Blocks"),(g,D)=>{const u=_("p-layout-default");return t(),a(u,{class:"blocks"},{header:c(()=>[f(o(k))]),default:c(()=>[r.value?(t(),i(h,{key:0},[l.value?(t(),a(o(y),{key:0})):(t(),a(o(B),{key:1,onDelete:o(e).refresh},null,8,["onDelete"]))],64)):x("",!0)]),_:1})}}});export{F as default};
//# sourceMappingURL=Blocks-pUCF37MF.js.map
