import{d as i,al as b,W as d,b as o,bk as _,c as f,e as c,a as v,f as D,u as h,o as B,g as u,h as n,ch as g,ci as C,aX as w}from"./index-BjFilUfY.js";import{u as x}from"./usePageTitle-J4-vayug.js";const S=i({__name:"BlockView",setup(y){const s=b(),a=h(),t=d("blockDocumentId"),l=o(()=>t.value?[t.value]:null),r=_(s.blockDocuments.getBlockDocument,l),e=o(()=>r.response),m=()=>{a.push(w.blocks())},k=o(()=>e.value?`Block: ${e.value.name}`:"Block");return x(k),(V,I)=>{const p=D("p-layout-default");return e.value?(B(),f(p,{key:0,class:"block-view"},{header:c(()=>[u(n(g),{"block-document":e.value,onDelete:m},null,8,["block-document"])]),default:c(()=>[u(n(C),{"block-document":e.value},null,8,["block-document"])]),_:1})):v("",!0)}}});export{S as default};
//# sourceMappingURL=BlockView-C0TdKoMt.js.map
