import{d as p,al as _,W as u,B as k,c as m,e as l,f as w,o as f,g as r,h as e,cN as i,cO as d}from"./index-BjFilUfY.js";import{u as P}from"./usePageTitle-J4-vayug.js";const x=p({__name:"WorkPoolEdit",async setup(h){let o,a;const s=_(),n=u("workPoolName"),t=([o,a]=k(()=>s.workPools.getWorkPoolByName(n.value)),o=await o,a(),o);return P("Create Work Pool"),(y,C)=>{const c=w("p-layout-default");return f(),m(c,null,{header:l(()=>[r(e(i),{"work-pool":e(t)},null,8,["work-pool"])]),default:l(()=>[r(e(d),{"work-pool":e(t)},null,8,["work-pool"])]),_:1})}}});export{x as default};
//# sourceMappingURL=WorkPoolEdit-C9oaSUdP.js.map
