import{d as m,al as _,t as i,b as s,c as a,e as n,f as k,o,g as f,h as t,cI as P,p as h,cJ as y,cK as g,F as v,a as w}from"./index-BjFilUfY.js";import{u as x}from"./usePageTitle-J4-vayug.js";const N=m({__name:"WorkPools",setup(C){const r=_(),c={interval:3e4},e=i(r.workPools.getWorkPools,[{}],c),l=s(()=>e.response??[]),p=s(()=>e.executed&&l.value.length==0),u=s(()=>e.executed);return x("Work Pools"),(b,B)=>{const d=k("p-layout-default");return o(),a(d,{class:"work-pools"},{header:n(()=>[f(t(P))]),default:n(()=>[u.value?(o(),h(v,{key:0},[p.value?(o(),a(t(y),{key:0})):(o(),a(t(g),{key:1,onUpdate:t(e).refresh},null,8,["onUpdate"]))],64)):w("",!0)]),_:1})}}});export{N as default};
//# sourceMappingURL=WorkPools-CTyaDqYq.js.map
