import{d as k,al as b,W as y,b as o,bk as i,c as _,e as l,a as m,f,o as T,g as t,h as c,ca as d,cb as g}from"./index-BjFilUfY.js";import{u as v}from"./usePageTitle-J4-vayug.js";const x=k({__name:"BlocksCatalogView",setup(B){const s=b(),a=y("blockTypeSlug"),n=o(()=>a.value?[a.value]:null),u=i(s.blockTypes.getBlockTypeBySlug,n),e=o(()=>u.response),p=o(()=>e.value?`Block Type: ${e.value.name}`:null);return v(p),(C,S)=>{const r=f("p-layout-default");return e.value?(T(),_(r,{key:0,class:"blocks-catalog-view"},{header:l(()=>[t(c(d),{"block-type":e.value},null,8,["block-type"])]),default:l(()=>[t(c(g),{"block-type":e.value},null,8,["block-type"])]),_:1})):m("",!0)}}});export{x as default};
//# sourceMappingURL=BlocksCatalogView-9hlADD_z.js.map
