import{d as g,al as h,W as w,b as u,am as R,b5 as D,t as I,c,e as n,f as i,u as S,ar as T,o as a,h as t,ct as B,a as r,cu as m,g as N,cv as V,a5 as W,aX as A}from"./index-BjFilUfY.js";import{u as E}from"./usePageTitle-J4-vayug.js";const X=g({__name:"ConcurrencyLimit",setup(O){const p=h(),y=w("concurrencyLimitId"),v=S(),d=u(()=>[{label:"Details",hidden:T.xl},{label:"Active Task Runs"}]),s=R("tab","Details"),{tabs:_}=D(d,s),b={interval:3e5},L=I(p.concurrencyLimits.getConcurrencyLimit,[y.value],b),e=u(()=>L.response);function f(){v.push(A.concurrencyLimits())}const C=u(()=>e.value?`Concurrency Limit: ${e.value.tag}`:"Concurrency Limit");return E(C),(P,l)=>{const k=i("p-tabs"),x=i("p-layout-well");return a(),c(x,{class:"concurrencyLimit"},{header:n(()=>[e.value?(a(),c(t(B),{key:0,"concurrency-limit":e.value,onDelete:f},null,8,["concurrency-limit"])):r("",!0)]),well:n(()=>[e.value?(a(),c(t(m),{key:0,alternate:"","concurrency-limit":e.value},null,8,["concurrency-limit"])):r("",!0)]),default:n(()=>[N(k,{selected:t(s),"onUpdate:selected":l[0]||(l[0]=o=>W(s)?s.value=o:null),tabs:t(_)},{details:n(()=>[e.value?(a(),c(t(m),{key:0,"concurrency-limit":e.value},null,8,["concurrency-limit"])):r("",!0)]),"active-task-runs":n(()=>{var o;return[(o=e.value)!=null&&o.activeSlots?(a(),c(t(V),{key:0,"active-slots":e.value.activeSlots},null,8,["active-slots"])):r("",!0)]}),_:1},8,["selected","tabs"])]),_:1})}}});export{X as default};
//# sourceMappingURL=ConcurrencyLimit-y8C9xyWo.js.map
