import{d as w,ay as y,az as v,al as C,t as F,b as c,aA as x,c as u,e as t,f as p,o as l,g as s,y as V,k as i,h as e,aB as D,aC as A,i as B,p as _,aD as R,aE as N,aF as S,aG as H,F as P,a as T,x as U,aH as E,l as I}from"./index-BjFilUfY.js";import{s as j,a as z}from"./index-Vv9yEoRy.js";const G={class:"workspace-dashboard__header-actions"},L={key:1,class:"workspace-dashboard__grid"},M={class:"workspace-dashboard__side"},K=w({__name:"Dashboard",setup(W){y(v,{interval:j(30)});const m=C(),r=F(m.flowRuns.getFlowRunsCount,[{}]),f=c(()=>r.executed),d=c(()=>r.response===0),k=[{text:"Dashboard"}],a=x({range:{type:"span",seconds:-z},tags:[]}),b=()=>I.map("WorkspaceDashboardFilter",a,"TaskRunsFilter");return(X,o)=>{const g=p("p-button"),h=p("p-layout-default");return l(),u(h,{class:"workspace-dashboard"},{header:t(()=>[s(e(B),{crumbs:k,class:"workspace-dashboard__page-heading"},V({_:2},[d.value?void 0:{name:"actions",fn:t(()=>[i("div",G,[s(e(D),{selected:e(a).tags,"onUpdate:selected":o[0]||(o[0]=n=>e(a).tags=n),"empty-message":"All tags"},null,8,["selected"]),s(e(A),{modelValue:e(a).range,"onUpdate:modelValue":o[1]||(o[1]=n=>e(a).range=n),class:"workspace-dashboard__date-select"},null,8,["modelValue"])])]),key:"0"}]),1024)]),default:t(()=>[f.value?(l(),_(P,{key:0},[d.value?(l(),u(e(R),{key:0})):(l(),_("div",L,[s(e(N),{filter:e(a)},null,8,["filter"]),i("div",M,[s(e(S),{filter:b}),s(e(H),{class:"workspace-dashboard__work-pools",filter:e(a)},null,8,["filter"])])]))],64)):T("",!0),s(e(E),{title:"Ready to scale?",subtitle:"Automations, role and object-level security, and serverless push work pools on Prefect Cloud"},{actions:t(()=>[s(g,{to:"https://www.prefect.io/cloud-vs-oss?utm_source=oss&utm_medium=oss&utm_campaign=oss&utm_term=none&utm_content=none",target:"_blank",primary:""},{default:t(()=>[U(" Upgrade to Cloud ")]),_:1})]),_:1})]),_:1})}}});export{K as default};
//# sourceMappingURL=Dashboard-BN5PMx2X.js.map
