import{b as a,b8 as c,h as n}from"./index-BjFilUfY.js";function i(...t){const r=[...t,"Prefect Server"],o=a(()=>r.map(e=>n(e)).filter(e=>e!==null).join(" • "));c(()=>document.title=o.value)}export{i as u};
//# sourceMappingURL=usePageTitle-J4-vayug.js.map
