import{d as c,al as k,W as t,B as _,c as w,e as r,f as i,o as P,g as n,h as o,cU as d,cV as f}from"./index-BjFilUfY.js";import{u as Q}from"./usePageTitle-J4-vayug.js";const x=c({__name:"WorkPoolQueueEdit",async setup(N){let e,u;const s=k(),a=t("workPoolName"),l=t("workPoolQueueName"),m=([e,u]=_(()=>s.workPoolQueues.getWorkPoolQueueByName(a.value,l.value)),e=await e,u(),e);return Q("Edit Work Pool Queue"),(h,y)=>{const p=i("p-layout-default");return P(),w(p,null,{header:r(()=>[n(o(d),{"work-pool-name":o(a),"work-pool-queue-name":o(l)},null,8,["work-pool-name","work-pool-queue-name"])]),default:r(()=>[n(o(f),{"work-pool-name":o(a),"work-pool-queue":o(m)},null,8,["work-pool-name","work-pool-queue"])]),_:1})}}});export{x as default};
//# sourceMappingURL=WorkPoolQueueEdit-vMSg2i_x.js.map
