import{d as m,al as y,W as _,t as k,b as s,c as e,e as o,f as v,A as u,o as t,h as c,au as d,a as r,aq as C,p as g,av as h,F as x}from"./index-BjFilUfY.js";import{u as A}from"./usePageTitle-J4-vayug.js";const T=m({__name:"ArtifactKey",setup(B){const n=y(),l=_("artifactKey"),i=k(n.artifacts.getArtifactCollection,[l]),a=s(()=>i.response),f=s(()=>a.value?`${u.info.artifact}: ${a.value.key}`:u.info.artifact);return A(f),(K,b)=>{const p=v("p-layout-default");return t(),e(p,{class:"artifact"},{header:o(()=>[a.value?(t(),e(c(d),{key:0,artifact:a.value},null,8,["artifact"])):r("",!0)]),default:o(()=>[a.value?(t(),e(c(C),{key:0,artifact:a.value},null,8,["artifact"])):r("",!0),a.value?(t(),g(x,{key:1},[a.value.key?(t(),e(c(h),{key:0,"artifact-key":a.value.key},null,8,["artifact-key"])):r("",!0)],64)):r("",!0)]),_:1})}}});export{T as default};
//# sourceMappingURL=ArtifactKey-BhAcJhcY.js.map
