import{d as o,c as r,e,f as c,o as n,g as a,h as t,aw as f,ax as p}from"./index-BjFilUfY.js";import{u}from"./usePageTitle-J4-vayug.js";const x=o({__name:"Artifacts",setup(_){return u("Artifacts"),(l,i)=>{const s=c("p-layout-default");return n(),r(s,{class:"artifacts"},{header:e(()=>[a(t(f))]),default:e(()=>[a(t(p))]),_:1})}}});export{x as default};
//# sourceMappingURL=Artifacts-DeSLpRKD.js.map
