import{d as m,al as _,t as i,b as s,c as o,e as n,f as b,o as t,g as w,h as a,bq as h,p as C,br as k,bs as y,F as v,a as x}from"./index-BjFilUfY.js";import{u as F}from"./usePageTitle-J4-vayug.js";const q=m({__name:"Flows",setup(g){const l=_(),c={interval:3e4},e=i(l.flows.getFlowsCount,[{}],c),r=s(()=>e.response??0),u=s(()=>e.executed&&r.value===0),p=s(()=>e.executed),d=()=>{e.refresh()};return F("Flows"),(B,D)=>{const f=b("p-layout-default");return t(),o(f,{class:"flows"},{header:n(()=>[w(a(h))]),default:n(()=>[p.value?(t(),C(v,{key:0},[u.value?(t(),o(a(k),{key:0})):(t(),o(a(y),{key:1,selectable:"",onDelete:d}))],64)):x("",!0)]),_:1})}}});export{q as default};
//# sourceMappingURL=Flows-DBjrOLtY.js.map
