import{d as e,c as o,h as a,o as t,cZ as r}from"./index-BjFilUfY.js";import{u as s}from"./usePageTitle-J4-vayug.js";const m=e({__name:"404",setup(n){return s("Page Not Found"),(c,p)=>(t(),o(a(r)))}});export{m as default};
//# sourceMappingURL=404-VDf6Ga34.js.map
