import{d as h,al as B,W as I,t as S,b as p,ae as T,am as V,c,e,f as r,A as _,an as $,o as l,h as t,ao as D,a as i,ap as d,p as W,g as s,aq as q,ar as E,as as v,x as H,m as L,at as m,a5 as P}from"./index-BjFilUfY.js";import{u as Q}from"./usePageTitle-J4-vayug.js";const U={key:0},J=h({__name:"Artifact",setup(j){const b=B(),k=I("artifactId"),w=S(b.artifacts.getArtifact,[k]),a=p(()=>w.response),n=T(!1),y=[{label:"Artifact"},{label:"Details"},{label:"Raw"}],u=V("tab","Artifact"),x=p(()=>a.value?`${_.info.artifact}: ${a.value.key??$(a.value.type)}`:_.info.artifact);return Q(x),(z,o)=>{const g=r("p-divider"),A=r("p-button"),C=r("p-content"),R=r("p-tabs"),N=r("p-layout-well");return l(),c(N,{class:"artifact"},{header:e(()=>[a.value?(l(),c(t(D),{key:0,artifact:a.value},null,8,["artifact"])):i("",!0)]),well:e(()=>[a.value?(l(),c(t(d),{key:0,artifact:a.value,alternate:""},null,8,["artifact"])):i("",!0)]),default:e(()=>[a.value?(l(),W("section",U,[s(t(q),{artifact:a.value},null,8,["artifact"]),s(g),t(E).xl?(l(),c(C,{key:0},{default:e(()=>[s(t(v),{artifact:a.value},null,8,["artifact"]),s(A,{class:"artifact__raw-data-button",small:"",onClick:o[0]||(o[0]=f=>n.value=!n.value)},{default:e(()=>[H(L(n.value?"Hide":"Show")+" raw data ",1)]),_:1}),n.value?(l(),c(t(m),{key:0,artifact:a.value},null,8,["artifact"])):i("",!0)]),_:1})):(l(),c(R,{key:1,selected:t(u),"onUpdate:selected":o[1]||(o[1]=f=>P(u)?u.value=f:null),tabs:y},{artifact:e(()=>[s(t(v),{artifact:a.value},null,8,["artifact"])]),details:e(()=>[s(t(d),{artifact:a.value},null,8,["artifact"])]),raw:e(()=>[s(t(m),{artifact:a.value},null,8,["artifact"])]),_:1},8,["selected"]))])):i("",!0)]),_:1})}}});export{J as default};
//# sourceMappingURL=Artifact-ClVhHhig.js.map
