import{d,ae as m,al as _,b as o,A as v,c as r,e as c,f as b,t as h,o as s,g as C,h as l,cB as k,p as y,cC as g,cD as x,F as B,a as V}from"./index-BjFilUfY.js";import{u as D}from"./usePageTitle-J4-vayug.js";const S=d({__name:"Variables",setup(E){const n=m(),a=()=>{var e;t.value.refresh(),(e=n.value)==null||e.refreshSubscriptions()},u=_(),t=o(()=>h(u.variables.getVariables)),p=o(()=>{var e;return t.value.executed&&((e=t.value.response)==null?void 0:e.length)===0}),i=o(()=>t.value.executed);return D(v.info.variables),(e,F)=>{const f=b("p-layout-default");return s(),r(f,{class:"variables"},{header:c(()=>[C(l(k),{onCreate:a})]),default:c(()=>[i.value?(s(),y(B,{key:0},[p.value?(s(),r(l(g),{key:0,onCreate:a})):(s(),r(l(x),{key:1,ref_key:"table",ref:n,onDelete:a,onUpdate:a},null,512))],64)):V("",!0)]),_:1})}}});export{S as default};
//# sourceMappingURL=Variables-Cq2h9b4d.js.map
