import{d as i,al as d,c as f,e as a,f as k,u as m,o as h,g as t,h as u,c0 as w,c1 as Q,aX as s,_ as c}from"./index-BjFilUfY.js";import{u as C}from"./usePageTitle-J4-vayug.js";const x=i({__name:"WorkQueueCreate",setup(g){const n=d(),o=m(),l=()=>{o.push(s.workQueues())},_=async r=>{try{const{id:e}=await n.workQueues.createWorkQueue(r);c("Work queue has been created","success"),o.push(s.workQueue(e))}catch(e){c("Error occurred while creating new work queue","error"),console.error(e)}};return C("Create Work Queue"),(r,e)=>{const p=k("p-layout-default");return h(),f(p,null,{header:a(()=>[t(u(w))]),default:a(()=>[t(u(Q),{action:"Create",onSubmit:_,onCancel:l})]),_:1})}}});export{x as default};
//# sourceMappingURL=WorkQueueCreate-BzUlVyIF.js.map
