import{d as k,al as y,ae as f,b as o,c7 as _,t as m,c as d,e as t,f as g,o as v,g as s,h as l,c8 as C,c9 as B}from"./index-BjFilUfY.js";import{u as T}from"./usePageTitle-J4-vayug.js";const w=k({__name:"BlocksCatalog",setup(h){const c=y(),a=f(null),p=o(()=>a.value?[a.value]:[]),{filter:n}=_({blockSchemas:{blockCapabilities:p}}),i=m(c.blockTypes.getBlockTypes,[n]),u=o(()=>i.response??[]);return T("Blocks Catalog"),(x,e)=>{const r=g("p-layout-default");return v(),d(r,{class:"blocks-catalog"},{header:t(()=>[s(l(C))]),default:t(()=>[s(l(B),{capability:a.value,"onUpdate:capability":e[0]||(e[0]=b=>a.value=b),"block-types":u.value},null,8,["capability","block-types"])]),_:1})}}});export{w as default};
//# sourceMappingURL=BlocksCatalog-DuImDk2y.js.map
