import{d as w,cw as m,am as C,c as h,e as a,f as u,o as y,g as l,h as o,i as r,cx as M,a5 as i,cy as x,cz as g,cA as G}from"./index-BjFilUfY.js";const T=w({__name:"ConcurrencyLimits",setup(R){const{showModal:t,open:p}=m(),{showModal:e,open:b}=m(),_=[{label:"Global"},{label:"Task Run"}],c=C("tab","Global");return(v,s)=>{const d=u("p-button"),f=u("p-tabs"),k=u("p-layout-default");return y(),h(k,{class:"concurrency-limits"},{header:a(()=>[l(o(r),{crumbs:[{text:"Concurrency"}]})]),default:a(()=>[l(f,{selected:o(c),"onUpdate:selected":s[2]||(s[2]=n=>i(c)?c.value=n:null),tabs:_},{global:a(()=>[l(o(r),{size:"lg",crumbs:[{text:"Global Concurrency Limits"}]},{"after-crumbs":a(()=>[l(d,{small:"",icon:"PlusIcon",onClick:o(p)},null,8,["onClick"])]),_:1}),l(o(M),{showModal:o(t),"onUpdate:showModal":s[0]||(s[0]=n=>i(t)?t.value=n:null)},null,8,["showModal"]),l(o(x),{class:"concurrency-limits__global-table"})]),"task-run":a(()=>[l(o(r),{size:"lg",crumbs:[{text:"Task Run Concurrency Limits"}]},{"after-crumbs":a(()=>[l(d,{small:"",icon:"PlusIcon",onClick:o(b)},null,8,["onClick"])]),_:1}),l(o(g),{showModal:o(e),"onUpdate:showModal":s[1]||(s[1]=n=>i(e)?e.value=n:null)},null,8,["showModal"]),l(o(G),{class:"concurrency-limits__task-limits-table"})]),_:1},8,["selected"])]),_:1})}}});export{T as default};
//# sourceMappingURL=ConcurrencyLimits-BpEizn2W.js.map
